import copy
import logging
from dataclasses import dataclass
from pprint import pformat

from panther_analysis_tool.analysis_utils import LoadAnalysisSpecsResult
from panther_analysis_tool.backend.client import Client as BackendClient
from panther_analysis_tool.backend.client import GenerateEnrichedEventParams
from panther_analysis_tool.constants import AnalysisTypes

TEST_CASE_FIELD_KEY_LOG = "Log"
TEST_CASE_FIELD_KEY_RESOURCE = "Resource"


@dataclass
class EventEnrichmentResult:
    original_test: dict
    enriched_test: dict

    def was_enriched(self) -> bool:
        return self.original_test != self.enriched_test


class EnrichedEventGenerator:
    """Enriches test data for analysis items."""

    def __init__(self, backend: BackendClient):
        """Initializes the TestDataEnricher.

        Args:
            backend: Backend API client.
        """
        self.backend = backend

    @staticmethod
    def _filter_analysis_items(
        analysis_items: list[LoadAnalysisSpecsResult],
    ) -> list[LoadAnalysisSpecsResult]:
        """Filters analysis items to only those that need test data enrichment.

        Args:
            analysis_items: A list of analysis items to filter.

        Returns:
            A list of analysis items that have the Tests property and an AnalysisType of RULE, POLICY, or SCHEDULED_RULE.
        """
        return [
            item
            for item in analysis_items
            if item.analysis_spec.get("Tests")
            and item.analysis_type()
            in [
                AnalysisTypes.RULE,
                AnalysisTypes.POLICY,
                AnalysisTypes.SCHEDULED_RULE,
                AnalysisTypes.CORRELATION_RULE,
            ]
        ]

    # pylint: disable=no-else-return
    @staticmethod
    def _convert_inline_json_dict_to_python_dict(data: dict) -> dict:
        """Converts YAML-loaded inline JSON to Python dictionaries. This allows them
        to be re-serialized into YAML instead of maintaining JSON formatting.
        """
        if isinstance(data, dict):
            new_dict = {}
            for key, value in data.items():
                if isinstance(value, (dict, list)):
                    new_dict[key] = EnrichedEventGenerator._convert_inline_json_dict_to_python_dict(
                        value  # type: ignore
                    )
                else:
                    new_dict[key] = value
            return new_dict
        elif isinstance(data, list):
            new_list = []
            for item in data:
                if isinstance(item, (dict, list)):
                    new_list.append(
                        EnrichedEventGenerator._convert_inline_json_dict_to_python_dict(item)
                    )
                else:
                    new_list.append(item)
            return new_list
        return data

    def _handle_analysis_item(
        self, analysis_id: str, test: dict, test_case_field_key: str
    ) -> EventEnrichmentResult:
        if test_case_field_key not in test:
            logging.error(
                "Skipping test case '%s' for %s, no event data found",
                test["Name"],
                analysis_id,
            )
            return EventEnrichmentResult(test, test)

        params = GenerateEnrichedEventParams(test[test_case_field_key])
        resp = self.backend.generate_enriched_event_input(params)

        if resp.status_code >= 400:
            logging.error(
                "Failed to enrich test data for %s: %s",
                analysis_id,
                resp.data,
            )
            return EventEnrichmentResult(test, test)

        logging.debug("GraphQL Response:\n%s", pformat(resp))

        enriched_test_data = resp.data.enriched_event

        # Create a copy of the existing test data so that we can compare before and after.
        enriched_test = copy.deepcopy(test)

        # We're only copying the p_enrichment field because it's the only net-new
        # field. This helps reduce unnecessary deserialize/serialize noise.
        #
        # If the returned "enriched event" has a p_enrichment field that's empty,
        # we'll just skip it. This reduces git diff noise.
        p_enrichment = enriched_test_data.get("p_enrichment", {})
        if p_enrichment in [{}, None]:
            logging.warning(
                "Skipping test case '%s' for %s, no enrichment data found",
                test["Name"],
                analysis_id,
            )
            # We're returning (test, enriched_test), but they're identical in content here.
            # We push enriched_test through _convert_inline_json_dict_to_python_dict so that
            # we can clean up the inline JSON content. Otherwise, it'll format as flow-style
            # inline JSON block in the YAML.
            return EventEnrichmentResult(
                test, EnrichedEventGenerator._convert_inline_json_dict_to_python_dict(enriched_test)
            )

        # Some test cases are pasted in as JSON. JSON does not roundtrip
        # nicely - often just rendering as one giant line after we add
        # the p_enrichment field.
        #
        # We're forcibly converting the test case data to a Python dict
        # so that it roundtrips out as YAML instead. This is noisy
        # for those tests that are in JSON format, but it's preferable
        # to the alternative.
        enriched_test[test_case_field_key] = enriched_test_data
        enriched_test = EnrichedEventGenerator._convert_inline_json_dict_to_python_dict(
            enriched_test
        )

        return EventEnrichmentResult(test, enriched_test)

    def _handle_rule_test(self, analysis_id: str, test: dict) -> EventEnrichmentResult:
        return self._handle_analysis_item(analysis_id, test, TEST_CASE_FIELD_KEY_LOG)

    def _handle_policy_test(self, analysis_id: str, test: dict) -> EventEnrichmentResult:
        return self._handle_analysis_item(analysis_id, test, TEST_CASE_FIELD_KEY_RESOURCE)

    def enrich_test_data(
        self, analysis_items: list[LoadAnalysisSpecsResult]
    ) -> list[LoadAnalysisSpecsResult]:
        """Enriches test data for analysis items.

        Args:
            analysis_items: A list of analysis items to enrich test data for.
        """
        logging.debug("Received %s analysis items", len(analysis_items))

        # Enrich any detections
        relevant_analysis_items = EnrichedEventGenerator._filter_analysis_items(analysis_items)
        logging.info(
            "Enriching test data for %s detections, after filtering", len(relevant_analysis_items)
        )

        enriched_analysis_items = []

        for analysis_item in relevant_analysis_items:
            analysis_id = analysis_item.analysis_spec.get(
                "RuleID"
            ) or analysis_item.analysis_spec.get("PolicyID")

            analysis_type = analysis_item.analysis_id()
            logging.info("Processing %s '%s'", analysis_type, analysis_id)
            tests = analysis_item.analysis_spec.get("Tests")

            results: list[EventEnrichmentResult] = []

            for test in tests:
                logging.info("Enriching test case '%s' for %s", test["Name"], analysis_id)
                if "Log" in test:
                    event_enrichment_result = self._handle_rule_test(analysis_id, test)
                elif "Resource" in test:
                    event_enrichment_result = self._handle_policy_test(analysis_id, test)
                else:
                    logging.warning(
                        "Skipping test case '%s' for %s, no event data found",
                        test["Name"],
                        analysis_id,
                    )
                    continue

                results.append(event_enrichment_result)

            logging.debug("Enrichment results:\n%s", pformat(results))

            if any(result.was_enriched() for result in results):
                analysis_item.analysis_spec["Tests"] = [result.enriched_test for result in results]
                analysis_item.serialize_to_file()
                enriched_analysis_items.append(analysis_item)
            else:
                logging.info("No test data enrichment available for rule '%s'", analysis_id)

        return enriched_analysis_items



import dataclasses
import io
from typing import Any

from ruamel import yaml
from textual.app import ComposeResult
from textual.containers import Horizontal
from textual.widget import Widget
from textual.widgets import DataTable, Label, TextArea

from panther_analysis_tool import analysis_utils
from panther_analysis_tool.core import parse

_EDITOR_THEME = "vscode_dark"

yaml_parser = yaml.YAML(typ="rt")
yaml_parser.preserve_quotes = True


@dataclasses.dataclass
class YamlDiffItem:
    key: str
    cust_val: Any
    panther_val: Any
    base_val: Any


@dataclasses.dataclass
class TableRow:
    type_label: str
    item_id_label: str
    description_label: str
    user_has_item: bool
    analysis_item: analysis_utils.AnalysisItem


class PythonWindow(TextArea):
    def __init__(self, text: str, read_only: bool = True, *args: Any, **kwargs: Any) -> None:
        super().__init__(
            text, theme=_EDITOR_THEME, language="python", read_only=read_only, *args, **kwargs
        )

    def on_mount(self) -> None:
        self.scroll_home(animate=False)


class YAMLWindow(TextArea):
    doc_lines: list[str] = []

    def __init__(self, text: str, read_only: bool = True, *args: Any, **kwargs: Any) -> None:
        super().__init__(
            text, theme=_EDITOR_THEME, language="yaml", read_only=read_only, *args, **kwargs
        )
        self.show_line_numbers = True
        self.highlight_cursor_line = True
        self.doc_lines = []

    def on_mount(self) -> None:
        for wrapped_line in self.wrapped_document.lines:
            for line in wrapped_line:
                self.doc_lines.append(line)

    def highlight_line(self, key: str) -> None:
        for i, line in enumerate(self.doc_lines):
            if line.strip().startswith(f"{key}:"):
                self.move_cursor((i, 0), center=True)
                break


class CustomerYAMLWindow(YAMLWindow):
    BORDER_TITLE = "Your YAML"


class PantherYAMLWindow(YAMLWindow):
    BORDER_TITLE = "Panther YAML"


class CustomerValueYAMLWindow(YAMLWindow):
    BORDER_TITLE = "Your Value [y]"


class PantherValueYAMLWindow(YAMLWindow):
    BORDER_TITLE = "Panther Value [p]"


class CustomerPythonWindow(PythonWindow):
    BORDER_TITLE = "Your Python"


@dataclasses.dataclass
class AnalysisDataTableItem:
    user_has_item: bool
    item: analysis_utils.AnalysisItem


class AnalysisItemDataTable(DataTable):
    all_table_data: list[TableRow]
    table_row_by_id: dict[str, TableRow]
    all_specs: list[AnalysisDataTableItem]
    current_rows: list[TableRow]
    CLONED_INDICATOR = "[green]Yes ✓[/green]"

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.add_columns("Cloned", "Type", "ID", "Description")
        self.all_table_data = []
        self.table_row_by_id = {}
        self.all_specs = []
        self.current_rows = []

    def add_specs_to_table(self, items: list[AnalysisDataTableItem]) -> None:
        """Add specs to the table."""
        self.clear()
        self.all_specs = items

        for item in items:
            row = self.table_item_to_table_row(item)
            self.all_table_data.append(row)
            self.add_row_to_table(row)
            self.table_row_by_id[item.item.analysis_id()] = row

    def table_item_to_table_row(self, item: AnalysisDataTableItem) -> TableRow:
        return TableRow(
            type_label=item.item.pretty_analysis_type().strip(),
            item_id_label=item.item.analysis_id().strip(),
            description_label=item.item.description().strip() or "(No description)",
            user_has_item=item.user_has_item,
            analysis_item=item.item,
        )

    def mark_user_has_item(self, item_id: str) -> None:
        self.table_row_by_id[item_id].user_has_item = True
        for spec in self.all_specs:
            if spec.item.analysis_id() == item_id:
                spec.user_has_item = True
                break

        # refresh data in table
        self.add_rows_to_table(self.current_rows)

    def add_row_to_table(self, row: TableRow) -> None:
        self.current_rows.append(row)
        status_indicator = self.CLONED_INDICATOR if row.user_has_item else "No"
        self.add_row(status_indicator, row.type_label, row.item_id_label, row.description_label)

    def clear(self, columns: bool = False) -> "AnalysisItemDataTable":
        self.current_rows = []
        return super().clear(columns=columns)

    def add_rows_to_table(self, rows: list[TableRow]) -> None:
        self.clear()
        for row in rows:
            self.add_row_to_table(row)

    def filter_by_filters(self, filters: list[parse.Filter]) -> None:
        self.clear()

        if not filters:
            self.reset_table()

        else:
            self.add_rows_to_table(
                [
                    row
                    for row in self.all_table_data
                    if analysis_utils.filters_match_analysis_item(filters, row.analysis_item)
                ]
            )

    def filter_by_id(self, _id: str) -> None:
        self.clear()
        for row in self.all_table_data:
            if _id == row.analysis_item.analysis_id():
                self.add_row_to_table(row)
                break

    def filter_by_type(self, _type: str) -> None:
        self.clear()
        for row in self.all_table_data:
            if (
                _type == row.analysis_item.pretty_analysis_type()
                or _type == row.analysis_item.analysis_type()
            ):
                self.add_row_to_table(row)

    def reset_table(self) -> None:
        self.clear()
        self.add_rows_to_table(self.all_table_data)


class DiffResolver(Widget):
    def __init__(self, diff_item: YamlDiffItem):
        super().__init__()
        self.diff_item = diff_item

    def fmt_panther_val(self) -> str:
        out = io.StringIO()
        yaml_parser.dump(self.diff_item.panther_val, out)
        return out.getvalue()

    def fmt_cust_val(self) -> str:
        out = io.StringIO()
        yaml_parser.dump(self.diff_item.cust_val, out)
        return out.getvalue()

    def compose(self) -> ComposeResult:
        yield Label(self.fmt_label())
        yield Horizontal(
            CustomerValueYAMLWindow(text="", id="customer-value-yaml"),
            PantherValueYAMLWindow(text="", id="panther-value-yaml"),
        )

    def fmt_label(self) -> str:
        return f'Resolving conflict for: {self.diff_item.key} (press "y" for your value or "p" for Panther\'s value)'



from typing import Any, List, Optional, Tuple

from panther_analysis_tool import analysis_utils
from panther_analysis_tool.core import (
    analysis_cache,
    install_item,
    parse,
    root,
    versions_file,
    yaml,
)


def run(analysis_id: Optional[str], filter_args: List[str]) -> Tuple[int, str]:
    try:
        root.chdir_to_project_root()
        analysis_cache.update_with_latest_panther_analysis(show_progress_bar=True)
        install(analysis_id, filter_args, show_installed_items=True)
    except (ValueError, analysis_cache.NoCacheException) as err:
        return 1, str(err)

    return 0, ""


def install(
    analysis_id: Optional[str], filter_args: List[str], show_installed_items: bool = False
) -> None:
    items_to_install = get_analysis_items(analysis_id, filter_args)

    if len(items_to_install) == 0:
        label = "analysis ID and filters"
        if analysis_id is None and len(filter_args) > 0:
            label = "filters"
        elif analysis_id is not None and len(filter_args) == 0:
            label = "analysis ID"

        raise ValueError(f"No items matched the {label}. Nothing to install.")

    install_analysis_items(items_to_install, show_installed_items=show_installed_items)


def get_analysis_items(
    analysis_id: Optional[str], filter_args: List[str]
) -> List[analysis_utils.AnalysisItem]:
    yaml_loader = yaml.BlockStyleYAML()
    cache = analysis_cache.AnalysisCache()
    versions = versions_file.get_versions().versions
    filters, filters_inverted = parse.parse_filter_args(filter_args)

    all_specs: List[analysis_utils.AnalysisItem] = []
    for _id in cache.list_spec_ids() if analysis_id is None else [analysis_id]:
        analysis_spec = cache.get_latest_spec(_id)
        if analysis_spec is None:
            continue
        if _id not in versions:
            continue

        loaded: dict[str, Any] = yaml_loader.load(analysis_spec.spec)

        if not analysis_utils.filter_analysis_spec(loaded, [*filters, *filters_inverted]):
            continue

        ver = versions[_id]
        all_specs.append(
            analysis_utils.AnalysisItem(
                yaml_file_contents=loaded,
                yaml_file_path=ver.history[ver.version].yaml_file_path,
                python_file_path=ver.history[ver.version].py_file_path,
                python_file_contents=cache.get_file_for_spec(
                    analysis_spec.id or -1, analysis_spec.version
                ),
            )
        )

    return all_specs


def install_analysis_items(
    items_to_install: List[analysis_utils.AnalysisItem], show_installed_items: bool = False
) -> None:
    for item in items_to_install:
        try:
            install_item.install_analysis_item(item, show_installed_items=show_installed_items)
        except FileExistsError:
            continue

    install_item.install_deps(items_to_install, show_installed_items=show_installed_items)



import dataclasses
import pathlib
from typing import Any

from panther_analysis_tool import analysis_utils
from panther_analysis_tool.constants import AnalysisTypes
from panther_analysis_tool.core import analysis_cache, parse, versions_file, yaml


@dataclasses.dataclass
class LoadedDataModelSpec:
    spec_item: analysis_cache.AnalysisSpec
    spec: dict[str, Any]
    log_types: list[str]


def install_analysis_item(
    item: analysis_utils.AnalysisItem, show_installed_items: bool = False
) -> None:
    """
    Installs the analysis item.

    Args:
        item: The analysis item to install.

    Raises:
        FileExistsError: If the analysis item already exists.
    """
    yaml_loader = yaml.BlockStyleYAML()

    yaml_path = pathlib.Path(item.yaml_file_path or "")
    if yaml_path.exists():
        raise FileExistsError(f"{item.analysis_id()} at {yaml_path} already exists")

    py_path = pathlib.Path(item.python_file_path or "")
    if item.python_file_path is not None and py_path.exists():
        raise FileExistsError(f"{item.analysis_id()} at {py_path} already exists")

    set_enabled_field(item.yaml_file_contents)
    set_base_version_field(item.yaml_file_contents)

    yaml_path.parent.mkdir(parents=True, exist_ok=True)
    with open(yaml_path, "wb") as yaml_file:
        yaml_loader.dump(item.yaml_file_contents, yaml_file)

    if item.python_file_path is not None and item.python_file_contents is not None:
        py_path.parent.mkdir(parents=True, exist_ok=True)
        with open(py_path, "wb") as py_file:
            py_file.write(item.python_file_contents)

    if show_installed_items:
        print(f"{item.pretty_analysis_type()} {item.analysis_id()} installed.")


def install_deps(
    items_with_deps: list[analysis_utils.AnalysisItem],
    show_installed_items: bool = False,
) -> None:
    """
    Installs the dependencies of the analysis item.

    Args:
        items_with_deps: The analysis items to install the dependencies for.
    """
    if len(items_with_deps) == 0:
        return

    all_log_types: set[str] = set()
    all_top_level_imports: set[str] = set()
    checked_global_helpers: set[str] = set()  # global helpers whose imports have been checked
    cache = analysis_cache.AnalysisCache()
    versions = versions_file.get_versions().versions
    global_helpers: dict[str, analysis_cache.AnalysisSpec] = {}  # filename -> spec
    data_models: list[LoadedDataModelSpec] = []

    for spec_id in cache.list_spec_ids():
        spec = cache.get_latest_spec(spec_id)
        if spec is None:
            continue

        # New loader per spec so roundtrip state is not shared across loads (see load_analysis_specs_ex).
        yaml_loader = yaml.BlockStyleYAML()
        loaded = yaml_loader.load(spec.spec)

        if loaded["AnalysisType"] == AnalysisTypes.GLOBAL:
            if "Filename" not in loaded:
                continue
            global_helpers[loaded["Filename"].split(".")[0]] = spec
        elif loaded["AnalysisType"] == AnalysisTypes.DATA_MODEL:
            data_models.append(
                LoadedDataModelSpec(
                    spec_item=spec,
                    spec=loaded,
                    log_types=loaded["LogTypes"] if "LogTypes" in loaded else [],
                )
            )

    # collect all imports from the items we are installing
    for item in items_with_deps:
        imports = parse.collect_top_level_imports(item.python_file_contents or b"")
        all_top_level_imports.update(imports)
        all_log_types.update(
            item.yaml_file_contents["LogTypes"] if "LogTypes" in item.yaml_file_contents else []
        )

    # collect all imports from the data models we are installing
    for log_type in all_log_types:
        for data_model in data_models:
            if log_type in data_model.log_types:
                python_file_content = cache.get_file_for_spec(
                    data_model.spec_item.id or -1, data_model.spec_item.version
                )
                if python_file_content is not None:
                    all_top_level_imports.update(
                        parse.collect_top_level_imports(python_file_content)
                    )

    # check all global helper imports
    check_global_helper_imports(
        all_top_level_imports, checked_global_helpers, global_helpers, cache
    )

    for import_ in all_top_level_imports:
        if import_ in global_helpers:
            try:
                item = cached_analysis_spec_to_analysis_item(
                    global_helpers[import_], cache, versions
                )
                install_analysis_item(item, show_installed_items=show_installed_items)
            except FileExistsError:
                pass

    # just LogTypes in data models, no ResourceTypes support
    for log_type in all_log_types:
        for data_model in data_models:
            if log_type in data_model.log_types:
                item = cached_analysis_spec_to_analysis_item(data_model.spec_item, cache, versions)
                set_enabled_field(item.yaml_file_contents)
                try:
                    install_analysis_item(item, show_installed_items=show_installed_items)
                except FileExistsError:
                    pass
                break


def check_global_helper_imports(
    all_top_level_imports: set[str],
    checked_global_helpers: set[str],
    global_helpers: dict[str, analysis_cache.AnalysisSpec],
    cache: analysis_cache.AnalysisCache,
) -> None:
    """
    Discovers transitive global helper dependencies by checking imports recursively.

    This function processes imports in needs_check and identifies which ones are global helpers.
    For each global helper found, it checks that helper's imports to discover additional
    transitive dependencies. All discovered global helper dependencies are added to needs_check.

    Args:
        all_top_level_imports: Set of import names to check. Modified in place to include all
            transitive global helper dependencies discovered during processing.
        checked_global_helpers: Set tracking which global helpers have already been
            processed. Modified in place to prevent duplicate processing.
        global_helpers: Dictionary mapping global helper filenames (without .py extension)
            to their AnalysisSpec objects.
        cache: AnalysisCache instance used to retrieve Python file contents for helpers.
    """
    # Use a queue-based approach to avoid recursion issues
    to_process = list(all_top_level_imports)

    while to_process:
        import_name = to_process.pop(0)

        # Skip if already checked or not a global helper
        if import_name in checked_global_helpers:
            continue
        if import_name not in global_helpers:
            continue

        # Get the helper's Python file
        helper = global_helpers[import_name]
        helper_py = cache.get_file_for_spec(helper.id or -1, helper.version)
        if helper_py is None:
            continue

        # Find imports in this helper
        helper_imports = parse.collect_top_level_imports(helper_py)

        # Mark this helper as checked
        checked_global_helpers.add(import_name)

        # Add any new global helper dependencies to the queue
        for helper_import in helper_imports:
            if helper_import in global_helpers and helper_import not in checked_global_helpers:
                all_top_level_imports.add(helper_import)
                if helper_import not in to_process:
                    to_process.append(helper_import)


def cached_analysis_spec_to_analysis_item(
    spec: analysis_cache.AnalysisSpec,
    cache: analysis_cache.AnalysisCache,
    versions: dict[str, versions_file.AnalysisVersionItem],
) -> analysis_utils.AnalysisItem:
    yaml_loader = yaml.BlockStyleYAML()
    loaded = yaml_loader.load(spec.spec)

    version_item = versions[spec.id_value]
    version_history_item = version_item.history[version_item.version]

    return analysis_utils.AnalysisItem(
        yaml_file_contents=loaded,
        yaml_file_path=version_history_item.yaml_file_path,
        raw_yaml_file_contents=spec.spec,
        python_file_path=version_history_item.py_file_path,
        python_file_contents=cache.get_file_for_spec(spec.id or -1, spec.version),
    )


def set_enabled_field(spec: dict[str, Any]) -> None:
    if spec["AnalysisType"] in [
        AnalysisTypes.RULE,
        AnalysisTypes.SCHEDULED_RULE,
        AnalysisTypes.CORRELATION_RULE,
        AnalysisTypes.POLICY,
        AnalysisTypes.DATA_MODEL,
        AnalysisTypes.LOOKUP_TABLE,
        AnalysisTypes.SAVED_QUERY,
        AnalysisTypes.SCHEDULED_QUERY,
        AnalysisTypes.DERIVED,
        AnalysisTypes.SIMPLE_DETECTION,
    ]:
        spec["Enabled"] = True


def set_base_version_field(spec: dict[str, Any]) -> None:
    spec["BaseVersion"] = versions_file.get_versions().get_current_version(
        analysis_utils.lookup_analysis_id(spec)
    )



import dataclasses
import io
import logging
import pathlib
import shutil
import tempfile

from panther_analysis_tool import analysis_utils
from panther_analysis_tool.constants import AutoAcceptOption
from panther_analysis_tool.core import diff, file_editor, git_helpers, yaml
from panther_analysis_tool.gui import yaml_conflict_resolver_gui


class MergeError(Exception):
    pass


@dataclasses.dataclass
class MergeableItem:
    user_item: analysis_utils.AnalysisItem
    latest_panther_item: analysis_utils.AnalysisItem
    base_panther_item: analysis_utils.AnalysisItem
    merged_item: analysis_utils.AnalysisItem | None = None
    latest_item_version: int = -1


def merge_item(
    mergeable_item: MergeableItem,
    solve_merge: bool,
    editor: str | None,
    auto_accept: AutoAcceptOption | None = None,
    write_merge_conflicts: bool = False,
) -> bool:
    # cannot write merge conflicts and auto accept at the same time
    # but write merge conflicts can be used with solve merge because that just means we are solving for one item at a time
    if write_merge_conflicts and auto_accept is not None:
        raise RuntimeError("Cannot write merge conflicts and auto accept at the same time")

    user_item = mergeable_item.user_item
    latest_item = mergeable_item.latest_panther_item
    base_item = mergeable_item.base_panther_item

    merged_yaml_contents = b""
    merged_python_contents = b""

    # merge yaml
    yaml_has_conflict, merged_yaml_contents = merge_file(
        solve_merge=solve_merge,
        # or b"" makes typing happy but it should never be None
        user=user_item.raw_yaml_file_contents or b"",
        base=base_item.raw_yaml_file_contents or b"",
        latest=latest_item.raw_yaml_file_contents or b"",
        user_python=b"",
        output_path=pathlib.Path(user_item.yaml_file_path or ""),
        editor=editor,
        auto_accept=auto_accept,
        write_merge_conflicts=write_merge_conflicts,
    )
    if yaml_has_conflict and not solve_merge and not write_merge_conflicts:
        # no need to merge python if yaml has a conflict because
        # we are just tracking what items have conflicts, not which files,
        # and yaml_has_conflict would be False if solve_merge is True
        return True

    # merge python (do this second since IDE may return before merge happens)
    py_has_conflict = False
    if user_item.python_file_contents is not None:
        py_has_conflict, merged_python_contents = merge_file(
            solve_merge=solve_merge,
            # or b"" makes typing happy but it should never be None
            user=user_item.python_file_contents or b"",
            base=base_item.python_file_contents or b"",
            latest=latest_item.python_file_contents or b"",
            user_python=user_item.python_file_contents or b"",
            output_path=pathlib.Path(user_item.python_file_path or ""),
            editor=editor,
            auto_accept=auto_accept,
            write_merge_conflicts=write_merge_conflicts,
        )
        if py_has_conflict and not write_merge_conflicts:
            return True

    if merged_yaml_contents != b"":
        pathlib.Path(user_item.yaml_file_path or "").write_bytes(merged_yaml_contents)

    if merged_python_contents != b"":
        pathlib.Path(user_item.python_file_path or "").write_bytes(merged_python_contents)

    if yaml_has_conflict or py_has_conflict:
        return True

    # the user item has been updated with the merged contents
    yaml_loader = yaml.BlockStyleYAML()
    raw_yaml_contents = pathlib.Path(user_item.yaml_file_path or "").read_bytes()
    mergeable_item.merged_item = analysis_utils.AnalysisItem(
        yaml_file_contents=yaml_loader.load(raw_yaml_contents),
        raw_yaml_file_contents=raw_yaml_contents,
        yaml_file_path=user_item.yaml_file_path,
        python_file_contents=(
            pathlib.Path(user_item.python_file_path).read_bytes()
            if user_item.python_file_path
            else None
        ),
        python_file_path=user_item.python_file_path,
    )

    return False


def merge_item_preview(
    mergeable_item: MergeableItem,
) -> bool:
    """
    Preview merging an analysis item without writing out the merged contents.

    Returns:
        True if there is a merge conflict, False otherwise.
    """
    user_item = mergeable_item.user_item
    latest_item = mergeable_item.latest_panther_item
    base_item = mergeable_item.base_panther_item

    # merge yaml
    yaml_has_conflict, _ = merge_file(
        solve_merge=False,
        # or b"" makes typing happy but it should never be None
        user=user_item.raw_yaml_file_contents or b"",
        base=base_item.raw_yaml_file_contents or b"",
        latest=latest_item.raw_yaml_file_contents or b"",
        user_python=b"",
        output_path=pathlib.Path(user_item.yaml_file_path or ""),
        editor=None,
    )
    if yaml_has_conflict:
        return True

    # merge python
    py_has_conflict = False
    if user_item.python_file_contents is not None:
        py_has_conflict, _ = merge_file(
            solve_merge=False,
            # or b"" makes typing happy but it should never be None
            user=user_item.python_file_contents or b"",
            base=base_item.python_file_contents or b"",
            latest=latest_item.python_file_contents or b"",
            user_python=user_item.python_file_contents or b"",
            output_path=pathlib.Path(user_item.python_file_path or ""),
            editor=None,
        )

    return py_has_conflict


# pylint: disable=too-many-locals,too-many-arguments
def merge_file(
    solve_merge: bool,
    user: bytes,
    base: bytes,
    latest: bytes,
    user_python: bytes,
    output_path: pathlib.Path,
    editor: str | None,
    auto_accept: AutoAcceptOption | None = None,
    write_merge_conflicts: bool = False,
) -> tuple[bool, bytes]:
    """
    Merge a file with git and solve the merge conflict if requested.
    YAML conflicts are solved with the yaml_conflict_resolver_gui.
    Python conflicts are solved with an editor.

    Args:
        solve_merge: Whether to solve the merge conflict.
        user: The user file contents.
        base: The base file contents.
        latest: The latest Panther file contents.
        user_python: The user Python file contents.
        output_path: The path to the output file.

    Returns:
        True if there was a merge conflict, False otherwise or if the user soled the merge conflict.
        The merged file contents.
    """
    ext = output_path.suffix
    yaml_loader = yaml.BlockStyleYAML()

    # this temp dir gets cleaned up automatically when it leaves the with block
    with tempfile.TemporaryDirectory(prefix="pat_merge_") as temp_dir_str:
        temp_dir = pathlib.Path(temp_dir_str)
        user_path = temp_dir / f"user{ext}"
        base_path = temp_dir / f"base{ext}"
        latest_path = temp_dir / f"latest{ext}"
        user_path.write_bytes(user)
        base_path.write_bytes(base)
        latest_path.write_bytes(latest)

        if ext in [".yml", ".yaml"]:
            user_python_path = temp_dir / "user_python.py"
            user_python_path.write_bytes(user_python)

            merge_dict = diff.Dict(yaml_loader.load(user_path), auto_accept=auto_accept)
            conflicts = merge_dict.merge_dict(
                yaml_loader.load(base_path), yaml_loader.load(latest_path)
            )

            if len(conflicts) == 0:
                out = io.BytesIO()
                yaml_loader.dump(merge_dict.customer_dict, out)
                return False, out.getvalue()  # merge was solved so no more conflict

            if write_merge_conflicts:
                # we cannot cleanly write the merge conflicts of the yaml diff, so we need to use git before writing the conflicts to the files
                return git_helpers.merge_file(
                    user_file_path=user_path,
                    base_file_path=base_path,
                    latest_file_path=latest_path,
                    auto_accept=auto_accept,
                )

            if not solve_merge:
                return True, b""

            app = yaml_conflict_resolver_gui.YAMLConflictResolverApp(
                customer_python=user_python.decode("utf-8"),
                raw_customer_yaml=user.decode("utf-8"),
                raw_panther_yaml=latest.decode("utf-8"),
                raw_base_yaml=base.decode("utf-8"),
                customer_dict=merge_dict.customer_dict,
                conflict_items=conflicts,
            )
            app.run()

            out = io.BytesIO()
            yaml_loader.dump(app.get_final_dict(), out)
            return False, out.getvalue()  # merge was solved so no more conflict

        # python merge
        has_conflict, merged_contents = git_helpers.merge_file(
            user_file_path=user_path,
            base_file_path=base_path,
            latest_file_path=latest_path,
            auto_accept=auto_accept,
        )
        if not has_conflict:
            return False, merged_contents

        if write_merge_conflicts:
            return True, merged_contents

        if not solve_merge:
            return True, b""

        # make a long-lived temp dir so that async editors can use the files in the directory.
        # IDE editing may outlive this function call because some IDEs are not blocking, like vscode and goland.
        # This dir will go in system temp and OS should eventually clean it up.
        long_lived_temp_dir = make_long_lived_temp_dir(temp_dir)

        merged_path = long_lived_temp_dir / f"merged{ext}"
        merged_path.write_bytes(merged_contents)

        async_edit = file_editor.merge_files_in_editor(
            file_editor.MergeableFiles(
                users_file=long_lived_temp_dir / user_path.name,
                base_file=long_lived_temp_dir / base_path.name,
                panthers_file=long_lived_temp_dir / latest_path.name,
                premerged_file=merged_path,
                output_file=output_path,
            ),
            editor=editor,
        )
        if not async_edit:
            # editing has finished so remove the long-lived temp dir
            shutil.rmtree(long_lived_temp_dir)

        return False, b""  # merge was solved so no more conflict


def make_long_lived_temp_dir(copy_dir: pathlib.Path) -> pathlib.Path:
    """
    Creates a long-lived temporary directory that is not cleaned up at exit but by the OS.
    Copy all files contents from the copy_dir to this directory for use.
    """
    temp_dir = tempfile.mkdtemp(prefix="pat_merge_")
    logging.debug("Long lived temp dir: %s", temp_dir)

    for file_path in copy_dir.iterdir():
        (pathlib.Path(temp_dir) / file_path.name).write_bytes(file_path.read_bytes())

    return pathlib.Path(temp_dir)



import enum
import importlib.metadata
import pathlib
from typing import Dict, Final

from schema import Schema

from panther_analysis_tool.schemas import (
    CORRELATION_RULE_SCHEMA,
    DATA_MODEL_SCHEMA,
    DERIVED_SCHEMA,
    GLOBAL_SCHEMA,
    LOOKUP_TABLE_SCHEMA,
    PACK_SCHEMA,
    POLICY_SCHEMA,
    RULE_SCHEMA,
    SAVED_QUERY_SCHEMA,
    SCHEDULED_QUERY_SCHEMA,
)

PACKAGE_NAME: Final = "panther_analysis_tool"

VERSION_STRING: Final = importlib.metadata.version(PACKAGE_NAME)

CONFIG_FILE = ".panther_settings.yml"
DATA_MODEL_LOCATION = "./data_models"
HELPERS_LOCATION = "./global_helpers"
DATA_MODEL_PATH_PATTERN = "*data_models*"
LUTS_PATH_PATTERN = "*lookup_tables*"
HELPERS_PATH_PATTERN = "*/global_helpers"
PACKS_PATH_PATTERN = "*/packs"
POLICIES_PATH_PATTERN = "*policies*"
QUERIES_PATH_PATTERN = "*queries*"
RULES_PATH_PATTERN = "*rules*"


class AnalysisTypes:
    DATA_MODEL = "datamodel"
    GLOBAL = "global"
    LOOKUP_TABLE = "lookup_table"
    PACK = "pack"
    POLICY = "policy"
    SAVED_QUERY = "saved_query"
    SCHEDULED_QUERY = "scheduled_query"
    RULE = "rule"
    DERIVED = "derived"
    SCHEDULED_RULE = "scheduled_rule"
    SIMPLE_DETECTION = "simple_detection"
    CORRELATION_RULE = "correlation_rule"


# The UserID is required by Panther for some API calls, but we have no way of
# acquiring it, and it isn't used for anything. This is a valid UUID used by the
# Panther deployment tool to indicate this action was performed automatically.
PANTHER_USER_ID = "00000000-0000-4000-8000-000000000000"

SCHEMAS: Dict[str, Schema] = {
    AnalysisTypes.DATA_MODEL: DATA_MODEL_SCHEMA,
    AnalysisTypes.GLOBAL: GLOBAL_SCHEMA,
    AnalysisTypes.LOOKUP_TABLE: LOOKUP_TABLE_SCHEMA,
    AnalysisTypes.PACK: PACK_SCHEMA,
    AnalysisTypes.POLICY: POLICY_SCHEMA,
    AnalysisTypes.SAVED_QUERY: SAVED_QUERY_SCHEMA,
    AnalysisTypes.SCHEDULED_QUERY: SCHEDULED_QUERY_SCHEMA,
    AnalysisTypes.RULE: RULE_SCHEMA,
    AnalysisTypes.DERIVED: DERIVED_SCHEMA,
    AnalysisTypes.SCHEDULED_RULE: RULE_SCHEMA,
    AnalysisTypes.CORRELATION_RULE: CORRELATION_RULE_SCHEMA,
}

SET_FIELDS = [
    "LogTypes",
    "PackIDs",
    "OutputIds",
    "SummaryAttributes",
    "Suppressions",
    "Tags",
]

BACKEND_FILTERS_ANALYSIS_SPEC_KEY = "_backend_filters"


class ReplayStatus:
    DONE = "DONE"
    CANCELED = "CANCELED"
    ERROR_EVALUATION = "ERROR_EVALUATION"
    ERROR_COMPUTATION = "ERROR_COMPUTATION"
    EVALUATION_IN_PROGRESS = "EVALUATION_IN_PROGRESS"
    COMPUTATION_IN_PROGRESS = "COMPUTATION_IN_PROGRESS"


ENABLE_CORRELATION_RULES_FLAG = "EnableCorrelationRules"

CACHE_DIR: Final = pathlib.Path(".cache")
PANTHER_ANALYSIS_SQLITE_FILE: Final = "panther-analysis.sqlite"
PANTHER_ANALYSIS_SQLITE_FILE_PATH: Final = CACHE_DIR / PANTHER_ANALYSIS_SQLITE_FILE
CACHED_VERSIONS_FILE_PATH: Final = CACHE_DIR / ".versions.yml"
LATEST_CACHED_PANTHER_ANALYSIS_FILE_PATH: Final = CACHE_DIR / "latest-panther-analysis.json"
CACHED_MIGRATION_STATUS_FILE_PATH: Final = CACHE_DIR / "migration_status.json"
MIGRATION_STATUS_FILE_PATH: Final = pathlib.Path("migration_status.md")
PAT_ROOT_FILE_NAME: Final = ".pat-root"

DEFAULT_EDITOR: Final = "vi"


class AutoAcceptOption(str, enum.Enum):
    YOURS = "yours"
    PANTHERS = "panthers"



from typing import Any, Literal

from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.containers import Horizontal
from textual.widgets import Footer, Label

from panther_analysis_tool.core import diff
from panther_analysis_tool.gui import widgets


class YAMLConflictResolverApp(App):
    """A Textual app to resolve YAML conflicts."""

    diff_items: list[widgets.YamlDiffItem]
    current_item_index: int = 0
    final_dict: dict[str, Any]
    customer_yaml: str
    panther_yaml: str
    customer_python: str

    BINDINGS = [
        Binding("ctrl+q", "quit", "Quit", show=True),
        Binding("p", "choose_panther", r"Select Panther's value", show=True),
        Binding("y", "choose_yours", r"Select your value", show=True),
        Binding("s", "switch_view", "Switch view", show=True),
    ]

    CSS_PATH = "yaml_conflict_resolver_gui.tcss"

    view: Literal["yaml", "python"] = "yaml"

    def __init__(
        self,
        customer_python: str,
        raw_customer_yaml: str,
        raw_panther_yaml: str,
        raw_base_yaml: str,
        customer_dict: dict,
        conflict_items: list[diff.DictMergeConflict],
    ) -> None:
        super().__init__()
        self.customer_python = customer_python
        self.customer_yaml = raw_customer_yaml
        self.panther_yaml = raw_panther_yaml
        self.final_dict = customer_dict
        self.diff_items = [
            widgets.YamlDiffItem(
                conflict.key, conflict.cust_val, conflict.latest_val, conflict.base_val
            )
            for conflict in conflict_items
        ]

    def compose(self) -> ComposeResult:
        yield Horizontal(
            widgets.CustomerPythonWindow(self.customer_python, id="customer-python"),
            widgets.CustomerYAMLWindow(self.customer_yaml, id="customer-yaml"),
            widgets.PantherYAMLWindow(self.panther_yaml, id="panther-yaml"),
            id="file-viewing",
        )
        yield widgets.DiffResolver(self.diff_items[0])
        yield Footer()

    def on_mount(self) -> None:
        self.update_view()
        self.update_diff_item()
        self.query_one("#customer-yaml").focus()

    def action_switch_view(self) -> None:
        self.view = "python" if self.view == "yaml" else "yaml"
        self.update_view()
        self.refresh()

    def update_view(self) -> None:
        if self.view == "yaml":
            self.query_one(widgets.CustomerPythonWindow).styles.display = "none"
            self.query_one(widgets.CustomerYAMLWindow).styles.display = "block"
            self.query_one(widgets.PantherYAMLWindow).styles.display = "block"
        elif self.view == "python":
            self.query_one(widgets.CustomerPythonWindow).styles.display = "block"
            self.query_one(widgets.CustomerYAMLWindow).styles.display = "none"
            self.query_one(widgets.PantherYAMLWindow).styles.display = "none"

    def action_choose_panther(self) -> None:
        self.update_final_dict(self.diff_items[self.current_item_index].panther_val)
        self.next_item()

    def action_choose_yours(self) -> None:
        self.update_final_dict(self.diff_items[self.current_item_index].cust_val)
        self.next_item()

    def update_final_dict(self, val: Any) -> None:
        key = self.diff_items[self.current_item_index].key
        if val is None:
            # Customer had removed this key; "yours" means keep it absent.
            self.final_dict.pop(key, None)
            return
        if isinstance(val, str):
            val = val.strip()
        self.final_dict[key] = val

    def next_item(self) -> None:
        self.current_item_index += 1
        if self.current_item_index >= len(self.diff_items):
            self.exit()
        else:
            self.update_diff_item()

    def update_diff_item(self) -> None:
        diff_resolver = self.query_one(widgets.DiffResolver)
        new_diff_item = self.diff_items[self.current_item_index]
        diff_resolver.diff_item = new_diff_item

        label = diff_resolver.query_one(Label)
        label.update(diff_resolver.fmt_label())

        panther_val = diff_resolver.fmt_panther_val()
        customer_val = diff_resolver.fmt_cust_val()

        diff_resolver.query_one(widgets.PantherValueYAMLWindow).text = panther_val
        diff_resolver.query_one(widgets.CustomerValueYAMLWindow).text = customer_val

        self.query_one("#panther-yaml", widgets.YAMLWindow).highlight_line(new_diff_item.key)
        self.query_one("#customer-yaml", widgets.YAMLWindow).highlight_line(new_diff_item.key)

        self.refresh()

    def get_final_dict(self) -> dict:
        return self.final_dict



import dataclasses
import importlib
import io
import json
import logging
import os
import pathlib
import re
import tempfile
from fnmatch import fnmatch
from importlib.abc import Loader
from typing import Any, Dict, Iterator, List, Optional, Tuple

import jsonschema
import schema
from jsonschema import Draft202012Validator
from ruamel.yaml import parser as YAMLParser
from ruamel.yaml import scanner as YAMLScanner

from panther_analysis_tool.backend.client import (
    BackendError,
)
from panther_analysis_tool.backend.client import Client as BackendClient
from panther_analysis_tool.backend.client import (
    GetRuleBodyParams,
    TestCorrelationRuleParams,
    TranspileFiltersParams,
    TranspileToPythonParams,
)
from panther_analysis_tool.constants import (
    BACKEND_FILTERS_ANALYSIS_SPEC_KEY,
    DATA_MODEL_LOCATION,
    DATA_MODEL_PATH_PATTERN,
    HELPERS_LOCATION,
    HELPERS_PATH_PATTERN,
    LUTS_PATH_PATTERN,
    PACKS_PATH_PATTERN,
    POLICIES_PATH_PATTERN,
    QUERIES_PATH_PATTERN,
    RULES_PATH_PATTERN,
    SCHEMAS,
    VERSION_STRING,
    AnalysisTypes,
)
from panther_analysis_tool.core import parse, yaml
from panther_analysis_tool.core.definitions import (
    ClassifiedAnalysis,
    ClassifiedAnalysisContainer,
    analysis_id_field_name,
)
from panther_analysis_tool.core.parse import Filter
from panther_analysis_tool.schemas import (
    ANALYSIS_CONFIG_SCHEMA,
    DERIVED_SCHEMA,
    POLICY_SCHEMA,
    RULE_SCHEMA,
    TYPE_SCHEMA,
)
from panther_analysis_tool.util import is_simple_detection
from panther_analysis_tool.validation import (
    contains_invalid_field_set,
    contains_invalid_table_names,
)


class AnalysisIDConflictException(Exception):
    """Exception for conflicting ids"""

    def __init__(self, analysis_id: str):
        self.message = f"Conflicting AnalysisID: [{analysis_id}]"
        super().__init__(self.message)


class AnalysisContainsDuplicatesException(Exception):
    """Exception for duplicate values in analysis specs"""

    def __init__(self, analysis_id: str, invalid_fields: List[str]):
        self.message = f'Specification file for [{analysis_id}] contains fields \
        with duplicate values: [{", ".join(x for x in invalid_fields)}]'
        super().__init__(self.message)


class AnalysisContainsInvalidTableNamesException(Exception):
    """Exception for invalid Panther Snowflake table names"""

    def __init__(self, analysis_id: str, invalid_table_names: List[str]):
        self.message = (
            f'Specification file for [{analysis_id}] contains invalid Panther table names: [{", ".join(x for x in invalid_table_names)}]. '
            'Try using a fully qualified table name such as "panther_logs.public.log_type" '
            "or setting --ignore-table-names for queries using non-Panther or non-Snowflake tables."
        )
        super().__init__(self.message)


def filter_analysis(
    analysis: List[ClassifiedAnalysis], filters: List[Filter], filters_inverted: List[Filter]
) -> List[ClassifiedAnalysis]:
    if filters is None:
        return analysis

    filtered_analysis = []
    for item in analysis:
        dir_name = item.dir_name
        file_name = item.file_name
        analysis_spec = item.analysis_spec
        if fnmatch(dir_name, HELPERS_PATH_PATTERN):
            logging.debug("auto-adding helpers file %s", os.path.join(file_name))
            filtered_analysis.append(ClassifiedAnalysis(file_name, dir_name, analysis_spec))
            continue
        if fnmatch(dir_name, DATA_MODEL_PATH_PATTERN):
            logging.debug("auto-adding data model file %s", os.path.join(file_name))
            filtered_analysis.append(ClassifiedAnalysis(file_name, dir_name, analysis_spec))
            continue

        if filter_analysis_spec(analysis_spec, [*filters, *filters_inverted]):
            filtered_analysis.append(ClassifiedAnalysis(file_name, dir_name, analysis_spec))
            continue

    return filtered_analysis


def filter_analysis_spec(analysis_spec: Dict[str, Any], filters: List[Filter]) -> bool:
    """
    Filters the analysis spec based on the filters and filters_inverted.
    Spec fields that match the filters are included, and spec fields that match the filters_inverted are excluded.
    Multiple filters are ANDed together, and multiple filter values in the same filter are ORed together.

    Args:
        analysis_spec: The analysis spec to filter.
        filters: The filters to apply.
        filters_inverted: The inverted filters to apply.

    Returns:
        True if the analysis spec matches the filters and filters_inverted, False otherwise.
    """
    for filt in filters:
        key, values = filt.key, filt.values
        spec_value = analysis_spec.get(key, "")
        spec_value = spec_value if isinstance(spec_value, list) else [spec_value]
        spec_value_normalized = {v.lower() if isinstance(v, str) else v for v in spec_value}
        values_normalized = {v.lower() if isinstance(v, str) else v for v in values}

        if not filt.inverted:
            if not spec_value_normalized.intersection(values_normalized):
                return False
        else:
            if spec_value_normalized.intersection(values_normalized):
                return False

    return True


def load_analysis_specs(
    directories: List[str], ignore_files: List[str]
) -> Iterator[Tuple[str, str, Any, Any]]:
    """Loads the analysis specifications from a file. It calls load_analysis_specs_ex
    and returns the spec_filename, relative_path, analysis_spec, and error to preserve
    historic callers.

    Args:
        directories: The relative path to Panther policies or rules.
        ignore_files: Files that Panther Analysis Tool should not process

    Yields:
        A tuple of the relative filepath, directory name, and loaded analysis specification dict.
    """
    for result in load_analysis_specs_ex(directories, ignore_files, roundtrip_yaml=False):
        yield result.spec_filename, result.relative_path, result.analysis_spec, result.error


def disable_all_base_detections(paths: List[str], ignore_files: List[str]) -> None:
    analysis_specs = list(load_analysis_specs_ex(paths, ignore_files, roundtrip_yaml=True))
    base_ids_to_disable = set()
    base_detection_key = "BaseDetection"
    rule_id_key = "RuleID"
    enabled_key = "Enabled"
    for analysis_spec_res in analysis_specs:
        spec: Dict[str, Any] = analysis_spec_res.analysis_spec
        base_id = spec.get(base_detection_key, "")
        if base_id == "":
            continue
        base_ids_to_disable.add(base_id)
    for base_detection_id in base_ids_to_disable:
        for analysis_spec_res in analysis_specs:
            rule: Dict[str, Any] = analysis_spec_res.analysis_spec
            if rule.get(rule_id_key, "") == base_detection_id:
                logging.info(
                    "Setting %s=False for %s", enabled_key, analysis_spec_res.spec_filename
                )
                rule[enabled_key] = False
                analysis_spec_res.serialize_to_file()


@dataclasses.dataclass
class LoadAnalysisSpecsResult:
    """
    The result of loading analysis specifications from a file.

    Attributes:
        spec_filename (str): Absolute path to the spec file.
        relative_path (str): Relative path to the spec file from the current working directory.
        analysis_spec (dict): The analysis specification.
        yaml_ctx (yaml.BlockStyleYAML): The YAML loader used to load the spec file.
        error (Exception | None): An error that occurred while loading the spec file.
        raw_spec_file_content (bytes | None): The raw content of the spec file.
    """

    spec_filename: str  # absolute path to the spec file
    relative_path: str
    analysis_spec: Any
    yaml_ctx: yaml.BlockStyleYAML
    error: Exception | None
    raw_spec_file_content: bytes | None

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, LoadAnalysisSpecsResult):
            return NotImplemented

        # skipping yaml_ctx because it's not relevant to equality of content
        # in analysis spec files
        same_spec_filename = self.spec_filename == other.spec_filename
        same_relative_path = self.relative_path == other.relative_path
        same_analysis_spec = self.analysis_spec == other.analysis_spec
        same_error = self.error == other.error

        return all(
            [
                same_spec_filename,
                same_relative_path,
                same_analysis_spec,
                same_error,
            ]
        )

    # pylint: disable=no-else-return
    def analysis_id(self) -> str:
        """Returns the analysis ID for this analysis spec."""
        return self.analysis_spec[self.analysis_id_field_name()]

    def analysis_type(self) -> str:
        """Returns the analysis type for this analysis spec."""
        return self.analysis_spec["AnalysisType"]

    def enabled(self) -> bool:
        """Returns True if the analysis spec is enabled, defaulting to True if no enabled field exists."""
        return self.analysis_spec.get("Enabled", True)

    def analysis_id_field_name(self) -> str:
        return analysis_id_field_name(self.analysis_type())

    # pylint: disable=line-too-long
    def __str__(self) -> str:
        return f"LoadAnalysisSpecsResult(spec_filename={self.spec_filename}, relative_path={self.relative_path}, analysis_spec={self.analysis_spec['AnalysisType']}, error={self.error})"

    def serialize_to_file(self) -> None:
        logging.debug("Writing analysis spec to %s", self.spec_filename)
        with open(self.spec_filename, "w", encoding="utf-8") as file:
            self.yaml_ctx.dump(self.analysis_spec, file)

    def python_file_path(self) -> pathlib.Path | None:
        """Returns the absolute path to the Python file for this analysis spec."""
        if "Filename" not in self.analysis_spec:
            return None

        path = pathlib.Path(self.spec_filename).parent / self.analysis_spec["Filename"]
        if not path.exists():
            return None

        return path

    def python_file_contents(self) -> bytes | None:
        """Returns the contents of the Python file for this analysis spec."""
        path = self.python_file_path()
        if path is None:
            return None

        return path.read_bytes()

    def pretty_analysis_type(self) -> str:
        return pretty_analysis_type(self.analysis_type(), plural=False)

    def unlink(self) -> None:
        pathlib.Path(self.spec_filename).unlink()
        py_file_path = self.python_file_path()
        if py_file_path is not None:
            pathlib.Path(py_file_path).unlink()

    def is_experimental(self) -> bool:
        return self.analysis_spec.get("Status", "").lower() == parse.EXPERIMENTAL_STATUS.lower()


# pylint: disable=too-many-locals
def load_analysis_specs_ex(
    directories: List[str], ignore_files: List[str], roundtrip_yaml: bool
) -> Iterator[LoadAnalysisSpecsResult]:
    """Loads the analysis specifications from a file. The _ex variant of this function returns
    a LoadAnalysisSpecsResult object that contains the yaml_ctx object that was used to load
    the specific YAML content. This allows us to roundtrip the YAML content back to the file with
    minimal changes.

    Args:
        directories: The relative path to Panther policies or rules.
        ignore_files: Files that Panther Analysis Tool should not process
        roundtrip_yaml: If True, roundtrip the YAML content back to the file with minimal changes. This
            is incompatible with most of PAT's YAML loading use-cases. However, if you need to modify
            YAML code, this is necessary.

    Yields:
        An instance of LoadAnalysisSpecsResult.
    """
    # setup a list of paths to ensure we do not import the same files
    # multiple times, which can happen when testing from root directory without filters
    ignored_normalized = ["dependabot.yml", "package.json", "package-lock.json"]
    for file in ignore_files:
        ignored_normalized.append(os.path.normpath(file))

    # Use safe mode when roundtrip is False, rt mode when roundtrip is True.
    # Create a fresh loader per file so each result has its own yaml_ctx. In roundtrip
    # mode a shared loader would accumulate state across loads and could corrupt
    # serialize_to_file() when multiple specs are loaded then some are written back.
    loaded_specs: List[Any] = []
    for directory in directories:
        for dirpath, dirnames, filenames in os.walk(directory):
            dirnames[:] = [
                dirname
                for dirname in dirnames
                if not dirname.startswith(".") and dirname != "__pycache__"
            ]

            # If the user runs with no path args, filter to make sure
            # we only run folders with valid analysis files. Ensure we test
            # files in the current directory by not skipping this iteration
            # when relative_path is the current dir
            if directory in [".", "./"] and dirpath not in [".", "./"]:
                if not any(
                    (
                        fnmatch(dirpath, path_pattern)
                        for path_pattern in (
                            DATA_MODEL_PATH_PATTERN,
                            HELPERS_PATH_PATTERN,
                            LUTS_PATH_PATTERN,
                            RULES_PATH_PATTERN,
                            PACKS_PATH_PATTERN,
                            POLICIES_PATH_PATTERN,
                            QUERIES_PATH_PATTERN,
                        )
                    )
                ):
                    logging.debug("Skipping path %s", dirpath)
                    continue
            for filename in sorted(filenames):
                # Skip hidden files
                if filename.startswith("."):
                    continue
                spec_filename = os.path.abspath(os.path.join(dirpath, filename))
                # skip loading files that have already been imported
                if spec_filename in loaded_specs:
                    continue
                # Dont load files that are explictly ignored
                relative_name = os.path.normpath(os.path.join(dirpath, filename))
                if relative_name in ignored_normalized:
                    logging.debug("ignoring file %s", relative_name)
                    continue
                loaded_specs.append(spec_filename)
                # New loader per file so roundtrip state is isolated (see doc above).
                if fnmatch(filename, "*.y*ml"):
                    yaml_loader = yaml.BlockStyleYAML(typ="rt" if roundtrip_yaml else "safe")
                    with open(spec_filename, "rb") as spec_file_obj:
                        try:
                            file_content = spec_file_obj.read()
                            yaml_content: dict[str, Any] | None = yaml_loader.load(
                                io.BytesIO(file_content)
                            )

                            if yaml_content is None or "AnalysisType" not in yaml_content:
                                continue  # skip files that aren't analysis items
                            yield LoadAnalysisSpecsResult(
                                spec_filename=spec_filename,
                                relative_path=dirpath,
                                analysis_spec=yaml_content,
                                yaml_ctx=yaml_loader,
                                error=None,
                                raw_spec_file_content=file_content,
                            )
                        except (YAMLParser.ParserError, YAMLScanner.ScannerError) as err:
                            # recreate the yaml object and yield the error
                            yield LoadAnalysisSpecsResult(
                                spec_filename=spec_filename,
                                relative_path=dirpath,
                                analysis_spec=None,
                                yaml_ctx=yaml_loader,
                                error=err,
                                raw_spec_file_content=None,
                            )


def to_relative_path(filename: str) -> str:
    cwd = os.getcwd()
    return os.path.relpath(filename, cwd)


# This function was generated in whole or in part by GitHub Copilot.
def get_simple_detections_as_python(
    specs: List[ClassifiedAnalysis], backend: Optional[BackendClient] = None
) -> List[ClassifiedAnalysis]:
    """Returns simple detections with transpiled Python."""
    enriched_specs = []
    if backend is not None:
        batch = [json.dumps(spec.analysis_spec, allow_nan=False) for spec in specs]
        try:
            params = TranspileToPythonParams(data=batch)
            response = backend.transpile_simple_detection_to_python(params)
            if response.status_code == 200:
                for i, result in enumerate(response.data.transpiled_python):
                    item = specs[i]
                    spec = item.analysis_spec
                    spec["body"] = result
                    enriched_specs.append(ClassifiedAnalysis(item.file_name, item.dir_name, spec))
            else:
                logging.warning(
                    "Error transpiling simple detection(s) to Python, skipping tests for simple detections."
                )
        except (BackendError, BaseException) as be_err:  # pylint: disable=broad-except
            logging.warning(
                "Error transpiling simple detection(s) to Python, skipping tests for simple detections:  %s",
                be_err,
            )
    else:
        logging.info("No backend client provided, skipping tests for simple detections.")
    return enriched_specs if enriched_specs else specs


def lookup_base_detection(the_id: str, backend: Optional[BackendClient] = None) -> Dict[str, Any]:
    """Attempts to lookup base detection via its id"""
    out: Dict[str, Any] = {}
    if backend is not None:
        try:
            params = GetRuleBodyParams(id=the_id)
            response = backend.get_rule_body(params)
            if response.status_code == 200:
                out["body"] = response.data.body
                out["tests"] = response.data.tests
            else:
                logging.warning(
                    "Unexpected error getting base detection, status code %s",
                    response.status_code,
                )
        except (BackendError, BaseException) as be_err:  # pylint: disable=broad-except
            logging.warning(
                "Error getting base detection %s: %s",
                the_id,
                be_err,
            )
    return out


def test_correlation_rule(
    spec: Dict[str, Any],
    backend: Optional[BackendClient] = None,
    test_names: Optional[List[str]] = None,
) -> List[Dict[str, Any]]:
    out: List[Dict[str, Any]] = []
    # dont make network call if there's no tests to run, or if no backend
    if "Tests" not in spec or backend is None:
        return out
    try:
        # Filter tests by name
        if test_names:
            tests = spec.get("Tests", [])
            tests = [t for t in tests if t["Name"] in test_names]
            spec["Tests"] = tests

        yaml_loader = yaml.BlockStyleYAML()
        string_io = io.StringIO()
        yaml_loader.dump(spec, stream=string_io)
        output_str = string_io.getvalue()
        string_io.close()
        resp = backend.test_correlation_rule(
            TestCorrelationRuleParams(
                yaml=output_str,
            )
        )
        out = resp.data.results
    except Exception as be_err:  # pylint: disable=broad-except
        logging.warning(
            "Error running tests remotely for correlation rule %s: %s",
            spec.get("RuleID", ""),
            be_err,
        )
    return out


def transpile_inline_filters(
    all_specs: ClassifiedAnalysisContainer, backend: Optional[BackendClient] = None
) -> None:
    """
    Transpiles all InlineFilters on detections from simple detection format to the backend format and saves the
    transpiled filter's JSON to the analysis spec under "_backend_filters".
    """
    inline_filters_key = "InlineFilters"

    # separate out all detections with inline filters
    # keep track of the detections without filters so the list can be used later
    all_detections_with_filters, detections, simple_detections = [], [], []
    for det in all_specs.detections:
        if inline_filters_key in det.analysis_spec:
            all_detections_with_filters.append(det)
        else:
            detections.append(det)
    for det in all_specs.simple_detections:
        if inline_filters_key in det.analysis_spec:
            all_detections_with_filters.append(det)
        else:
            simple_detections.append(det)

    # if there are no InlineFilters at all then we can just return
    if len(all_detections_with_filters) == 0:
        return

    if backend is not None:
        batch = [
            json.dumps(d.analysis_spec.get("InlineFilters"), allow_nan=False)
            for d in all_detections_with_filters
        ]
        try:
            params = TranspileFiltersParams(data=batch, pat_version=VERSION_STRING)
            response = backend.transpile_filters(params)

            if response.status_code == 200:
                # set the translated backend filters on the analysis spec
                for i, result in enumerate(response.data.transpiled_filters):
                    all_detections_with_filters[i].analysis_spec[
                        BACKEND_FILTERS_ANALYSIS_SPEC_KEY
                    ] = json.loads(result)
                # separate the simple detections from the other detections
                for det in all_detections_with_filters:
                    if is_simple_detection(det.analysis_spec):
                        simple_detections.append(det)
                    else:
                        detections.append(det)
                # replace the lists in all specs with the new specs
                all_specs.detections = detections
                all_specs.simple_detections = simple_detections
            else:
                logging.warning(
                    "Error transpiling InlineFilter(s), skipping InlineFilters during testing"
                )
        except (BackendError, BaseException) as be_err:  # pylint: disable=broad-except
            logging.warning(
                "Error transpiling InlineFilter(s), skipping InlineFilters during testing:  %s",
                be_err,
            )
    else:
        logging.info("No backend client provided, skipping InlineFilters during testing")


def load_analysis(
    path: str,
    ignore_table_names: bool,
    valid_table_names: List[str],
    ignore_files: List[str],
    ignore_extra_keys: bool,
) -> Tuple[ClassifiedAnalysisContainer, List[Any]]:
    """Loads each policy or rule into memory.

    Args:
        path: path to root folder with rules and policies
        ignore_table_names: validate or ignore table names
        valid_table_names: list of valid table names, other will be treated as invalid
        ignore_files: Files that Panther Analysis Tool should not process

    Returns:
        A tuple of the valid and invalid rules and policies
    """
    search_directories = [path]
    for directory in (
        HELPERS_LOCATION,
        "." + HELPERS_LOCATION,  # Try the parent directory as well
        DATA_MODEL_LOCATION,
        "." + DATA_MODEL_LOCATION,  # Try the parent directory as well
    ):
        absolute_dir_path = os.path.abspath(os.path.join(path, directory))
        absolute_helper_path = os.path.abspath(directory)

        if os.path.exists(absolute_dir_path):
            search_directories.append(absolute_dir_path)
        if os.path.exists(absolute_helper_path):
            search_directories.append(absolute_helper_path)

    # First classify each file, always include globals and data models location
    specs, invalid_specs = classify_analysis(
        list(load_analysis_specs(search_directories, ignore_files)),
        ignore_table_names=ignore_table_names,
        valid_table_names=valid_table_names,
        ignore_extra_keys=ignore_extra_keys,
    )

    return specs, invalid_specs


# pylint: disable=too-many-locals,too-many-statements
def classify_analysis(
    specs: List[Tuple[str, str, Any, Any]],
    ignore_table_names: bool,
    valid_table_names: List[str],
    ignore_extra_keys: bool,
) -> Tuple[ClassifiedAnalysisContainer, List[Any]]:
    # First setup return dict containing different
    # types of detections, meta types that can be zipped
    # or uploaded
    all_specs = ClassifiedAnalysisContainer()

    invalid_specs: List[Tuple[str, Exception]] = []
    # each analysis type must have a unique id, track used ids and
    # add any duplicates to the invalid_specs
    analysis_ids: List[str] = []

    # Create a json validator and check the schema only once rather than during every loop
    json_validator = Draft202012Validator(ANALYSIS_CONFIG_SCHEMA)

    # pylint: disable=too-many-nested-blocks
    for analysis_spec_filename, dir_name, analysis_spec, error in specs:
        keys: List[Any] = []
        tmp_logtypes: Any = None
        tmp_logtypes_key: Any = None
        try:
            # check for parsing errors from json.loads (ValueError) / yaml.safe_load (YAMLError)
            if error:
                raise error
            # validate the schema has a valid analysis type
            TYPE_SCHEMA.validate(analysis_spec)
            analysis_type = analysis_spec["AnalysisType"]
            if analysis_spec.get("BaseDetection"):
                analysis_schema = SCHEMAS["derived"]
            else:
                analysis_schema = SCHEMAS[analysis_type]
            keys = list(analysis_schema.schema.keys())
            # Special case for ScheduledQueries to only validate the types
            if "ScheduledQueries" in analysis_spec:
                for each_key in analysis_schema.schema.keys():
                    if str(each_key) == "Or('LogTypes', 'ScheduledQueries')":
                        tmp_logtypes_key = each_key
                        break
                if not tmp_logtypes:
                    tmp_logtypes = analysis_schema.schema[tmp_logtypes_key]
                analysis_schema.schema[tmp_logtypes_key] = [str]

            if analysis_schema in [RULE_SCHEMA, POLICY_SCHEMA, DERIVED_SCHEMA]:
                analysis_schema._ignore_extra_keys = (  # pylint: disable=protected-access
                    ignore_extra_keys
                )
            analysis_schema.validate(analysis_spec)

            # lookup the analysis type id and validate there aren't any conflicts
            analysis_id = lookup_analysis_id(analysis_spec)
            if analysis_id in analysis_ids:
                raise AnalysisIDConflictException(analysis_id)
            # check for duplicates where panther expects a unique set
            invalid_fields = contains_invalid_field_set(analysis_spec)
            if invalid_fields:
                raise AnalysisContainsDuplicatesException(analysis_id, invalid_fields)
            if analysis_type == AnalysisTypes.SCHEDULED_QUERY and not ignore_table_names:
                invalid_table_names = contains_invalid_table_names(
                    analysis_spec, analysis_id, valid_table_names
                )
                if invalid_table_names:
                    raise AnalysisContainsInvalidTableNamesException(
                        analysis_id, invalid_table_names
                    )
            analysis_ids.append(analysis_id)

            # Raise warnings for dedup minutes
            if "DedupPeriodMinutes" in analysis_spec:
                if analysis_spec["DedupPeriodMinutes"] == 0:
                    msg = (
                        f"DedupPeriodMinutes is set to 0 for {analysis_id}. "
                        "This will be ignored by the backend. "
                        "If you want to disable dedup, "
                        "alter the `dedup` function to return 'p_row_id' instead."
                    )
                    logging.warning(msg)
                elif analysis_spec["DedupPeriodMinutes"] < 5:
                    msg = (
                        f"DedupPeriodMinutes for {analysis_id} is less than 5. "
                        "This is below Panther's DedupPeriodMinutes threshold, "
                        "and will be treated as '5' upon upload."
                    )
                    logging.warning(msg)

            classified_analysis = ClassifiedAnalysis(
                analysis_spec_filename, dir_name, analysis_spec
            )

            json_validator.validate(analysis_spec)

            all_specs.add_classified_analysis(analysis_type, classified_analysis)

        except schema.SchemaWrongKeyError as err:
            invalid_specs.append((analysis_spec_filename, handle_wrong_key_error(err, keys)))
        except (
            schema.SchemaMissingKeyError,
            schema.SchemaForbiddenKeyError,
            schema.SchemaUnexpectedTypeError,
        ) as err:
            invalid_specs.append((analysis_spec_filename, err))
            continue
        except schema.SchemaError as err:
            # Intercept the error, otherwise the error message becomes confusing and unreadable
            error = err
            err_str = str(err)
            first_half = err_str.split(":", maxsplit=1)[0]
            second_half = err_str.split(")", maxsplit=1)[-1]
            if "LogTypes" in str(err):
                error = schema.SchemaError(f"{first_half}: LOG_TYPE_REGEX{second_half}")
            elif "ResourceTypes" in str(err):
                error = schema.SchemaError(f"{first_half}: RESOURCE_TYPE_REGEX{second_half}")
            invalid_specs.append((analysis_spec_filename, error))
        except jsonschema.exceptions.ValidationError as err:
            error_message = f"{getattr(err, 'json_path', 'error')}: {err.message}"
            invalid_specs.append(
                (
                    analysis_spec_filename,
                    jsonschema.exceptions.ValidationError(error_message),
                )
            )
        except Exception as err:  # pylint: disable=broad-except
            # Catch arbitrary exceptions thrown by bad specification files
            invalid_specs.append((analysis_spec_filename, err))
            continue
        finally:
            # Restore original values
            if tmp_logtypes and tmp_logtypes_key:
                analysis_schema.schema[tmp_logtypes_key] = tmp_logtypes

    return all_specs, invalid_specs


def handle_wrong_key_error(err: schema.SchemaWrongKeyError, keys: list) -> Exception:
    regex = r"Wrong key(?:s)? (.+?) in (.*)$"
    matches = re.match(regex, str(err))
    msg = "{} not in list of valid keys: {}"
    try:
        if matches:
            raise schema.SchemaWrongKeyError(msg.format(matches.group(1), keys)) from err
        raise schema.SchemaWrongKeyError(msg.format("UNKNOWN_KEY", keys)) from err
    except schema.SchemaWrongKeyError as exc:
        return exc


def lookup_analysis_id(analysis_spec: Any) -> str:
    """Returns the analysis ID for a given analysis spec."""
    id_field = analysis_id_field_name(analysis_spec["AnalysisType"])
    return analysis_spec.get(id_field, "UNKNOWN_ID")


def load_module(filename: str) -> Tuple[Any, Any]:
    """Loads the analysis function module from a file.

    Args:
        filename: The relative path to the file.

    Returns:
        A loaded Python module.
    """
    module_name = filename.split(".")[0]
    spec = importlib.util.spec_from_file_location(module_name, filename)
    module = importlib.util.module_from_spec(spec)  # type: ignore
    try:
        assert isinstance(spec.loader, Loader)  # type: ignore # nosec
        spec.loader.exec_module(module)  # type: ignore
    except FileNotFoundError as err:
        print("\t[ERROR] File not found: " + filename + ", skipping\n")
        return None, err
    except Exception as err:  # pylint: disable=broad-except
        # Catch arbitrary exceptions thrown by user code
        print("\t[ERROR] Error loading module, skipping\n")
        return None, err
    return module, None


def get_tmp_helper_module_location() -> str:
    return os.path.join(tempfile.gettempdir(), "panther-path", "globals")


@dataclasses.dataclass
class AnalysisItem:
    yaml_file_contents: dict[str, Any]
    raw_yaml_file_contents: Optional[bytes] = None
    yaml_file_path: Optional[str] = None
    python_file_contents: Optional[bytes] = None
    python_file_path: Optional[str] = None

    def __eq__(self, value: object) -> bool:
        return (
            isinstance(value, AnalysisItem)
            and self.yaml_file_contents == value.yaml_file_contents
            and self.yaml_file_path == value.yaml_file_path
            and self.python_file_path == value.python_file_path
            and self.python_file_contents == value.python_file_contents
            # not including raw_yaml_file_contents because whitespace differences and such don't matter
            # and yaml_file_contents is the only thing that matters for equality
        )

    def analysis_type(self) -> str:
        return self.yaml_file_contents["AnalysisType"]

    def analysis_id(self) -> str:
        return lookup_analysis_id(self.yaml_file_contents)

    def description(self) -> str:
        return self.yaml_file_contents.get("Description", "")

    def pretty_analysis_type(self, plural: bool = False) -> str:
        return pretty_analysis_type(self.analysis_type(), plural)

    def empty(self) -> bool:
        return (
            self.yaml_file_contents == {}
            and (self.raw_yaml_file_contents is None or self.raw_yaml_file_contents == b"{}")
            and (self.python_file_contents is None or self.python_file_contents == b"")
        )

    def analysis_id_field_name(self) -> str:
        return analysis_id_field_name(self.analysis_type())


# pylint: disable=too-many-return-statements
def pretty_analysis_type(analysis_type: str, plural: bool = False) -> str:
    match analysis_type:
        case AnalysisTypes.RULE:
            return "Rule" if not plural else "Rules"
        case AnalysisTypes.CORRELATION_RULE:
            return "Correlation Rule" if not plural else "Correlation Rules"
        case AnalysisTypes.SCHEDULED_RULE:
            return "Scheduled Rule" if not plural else "Scheduled Rules"
        case AnalysisTypes.DATA_MODEL:
            return "Data Model" if not plural else "Data Models"
        case AnalysisTypes.POLICY:
            return "Policy" if not plural else "Policies"
        case AnalysisTypes.GLOBAL:
            return "Global Helper" if not plural else "Global Helpers"
        case AnalysisTypes.SCHEDULED_QUERY:
            return "Scheduled Query" if not plural else "Scheduled Queries"
        case AnalysisTypes.SAVED_QUERY:
            return "Saved Query" if not plural else "Saved Queries"
        case AnalysisTypes.PACK:
            return "Pack" if not plural else "Packs"
        case AnalysisTypes.LOOKUP_TABLE:
            return "Lookup Table" if not plural else "Lookup Tables"
        case AnalysisTypes.DERIVED:
            return "Derived Detection" if not plural else "Derived Detections"
        case AnalysisTypes.SIMPLE_DETECTION:
            return "Simple Detection" if not plural else "Simple Detections"
        case _:
            raise ValueError(f"Unsupported analysis type: {analysis_type}")


def filters_match_analysis_item(filters: list[parse.Filter], item: AnalysisItem) -> bool:
    """
    Check if an analysis item matches all of the provided filters.

    This function handles two types of filters:
    1. **Regular filters** (with a non-empty key): These are structured filters like
       `RuleID=AWS.S3.Bucket.PublicRead` or `Severity=Critical`. They are matched
       against the item's YAML specification using `filter_analysis_spec`.
    2. **Text filters** (with an empty key): These are plain text search terms that
       must appear in EITHER the item's YAML content OR Python content.

    An item matches only if it satisfies ALL filters. If any filter
    fails to match, the function returns False.

    Args:
        filters: List of Filter objects to match against. Filters with empty keys
            are treated as text search filters, while filters with keys are treated
            as structured filters.
        item: The analysis item to check against the filters.

    Returns:
        True if the item matches all filters, False otherwise.

    Examples:
        >>> filters = [
        ...     Filter(key="RuleID", values=["AWS.S3.Bucket.PublicRead"]),
        ...     Filter(key="Severity", values=["Critical"]),
        ...     Filter(key="", values=["def rule"])
        ... ]
        >>> # Returns True only if:
        >>> # - RuleID is "AWS.S3.Bucket.PublicRead" AND
        >>> # - Severity is "Critical" AND
        >>> # - "def rule" appears in the YAML OR Python
    """
    raw_yaml = item.raw_yaml_file_contents.decode("utf-8") if item.raw_yaml_file_contents else ""
    python = item.python_file_contents.decode("utf-8") if item.python_file_contents else ""

    reg_filters = [filt for filt in filters if filt.key != ""]
    text_filters = [filt for filt in filters if filt.key == ""]

    for filt in text_filters:
        if filt.values[0] not in raw_yaml and filt.values[0] not in python:
            return False

    if not filter_analysis_spec(item.yaml_file_contents, reg_filters):
        return False

    return True



import datetime
import io
import sys
import zipfile
from dataclasses import dataclass
from statistics import mean, median
from typing import List, Optional, Tuple, Union

import dateutil.parser

from panther_analysis_tool.analysis_utils import ClassifiedAnalysis
from panther_analysis_tool.backend.client import Client as BackendClient
from panther_analysis_tool.backend.client import MetricsParams, PerfTestParams
from panther_analysis_tool.constants import AnalysisTypes, ReplayStatus
from panther_analysis_tool.core.parse import Filter
from panther_analysis_tool.util import log_and_write_to_file
from panther_analysis_tool.zip_chunker import (
    ZipArgs,
    analysis_for_chunks,
    chunk_analysis,
)


class PerformanceTestIteration:
    def __init__(self, read_time_nanos: int, processing_time_nanos: int) -> None:
        self.read_time_nanos = read_time_nanos
        self.processing_time_nanos = processing_time_nanos

    def __str__(self) -> str:
        return (
            f"Read time (nanoseconds): {self.read_time_nanos}\n"
            f"Processing time (nanoseconds): {self.processing_time_nanos}"
        )


# pylint: disable=too-many-instance-attributes
@dataclass
class BenchmarkArgs:
    out: str
    path: str
    ignore_files: List[str]
    filters: List[Filter]
    filters_inverted: List[Filter]
    iterations: int
    log_type: Optional[str]
    hour: Optional[datetime.datetime]


def run(  # pylint: disable=too-many-locals
    backend: BackendClient, args: BenchmarkArgs
) -> Tuple[int, str]:
    if backend is None or not backend.supports_perf_test():
        return 1, "Invalid backend. `benchmark` is only supported via API token"

    if args.iterations <= 0:
        return 1, f"benchmark must perform at least 1 iteration, {args.iterations} requested"

    zip_args = ZipArgs(
        out=args.out,
        path=args.path,
        ignore_files=args.ignore_files,
        filters=args.filters,
        filters_inverted=args.filters_inverted,
    )
    analyses = analysis_for_chunks(zip_args, no_helpers=True)

    rule_or_err = validate_rule_count(analyses)
    if isinstance(rule_or_err, str):
        return 1, rule_or_err

    log_type, err_msg = validate_log_type(args.log_type, rule_or_err)
    if err_msg is not None or not isinstance(log_type, str):
        return 1, err_msg or "No log_type found"

    hour_or_err = validate_hour(args.hour, log_type, backend)
    if isinstance(hour_or_err, str):
        return 1, hour_or_err

    chunks = chunk_analysis(analyses)
    buffer = io.BytesIO()
    with zipfile.ZipFile(buffer, "w", zipfile.ZIP_DEFLATED) as zip_out:
        for name in chunks[0].files:
            zip_out.write(name)
    buffer.seek(0, 0)

    now = datetime.datetime.now()
    timeout = now + datetime.timedelta(minutes=13)

    params = PerfTestParams(
        zip_bytes=buffer.read(),
        log_type=log_type,
        hour=hour_or_err,
        timeout=timeout.astimezone(),
    )
    iterations: List[PerformanceTestIteration] = []
    logged = False
    for _ in range(0, args.iterations):
        replay_response = backend.run_perf_test(params)
        if replay_response.data.state == ReplayStatus.CANCELED:
            if len(iterations) == 0:
                log_extreme_timeout(args.out, hour_or_err, now)
                logged = True
            break
        if replay_response.data.state in [
            ReplayStatus.ERROR_EVALUATION,
            ReplayStatus.ERROR_COMPUTATION,
        ]:
            log_error(args.out, hour_or_err, now)
            if len(iterations) == 0:
                logged = True
            break
        iterations.append(
            PerformanceTestIteration(
                read_time_nanos=replay_response.data.replay_summary.read_time_nanos,
                processing_time_nanos=replay_response.data.replay_summary.processing_time_nanos,
            )
        )

    if not logged:
        log_output(args.out, hour_or_err, iterations, rule_or_err, now)

    return 0, ""


def validate_rule_count(analyses: List[ClassifiedAnalysis]) -> Union[ClassifiedAnalysis, str]:
    if len(analyses) != 1:
        return (
            f"Only 1 detection must be specified for benchmarking, but {len(analyses)} were specified. The rule must"
            f" be the only item in the working directory or specified by --path, --ignore-files, and/or --filter. The"
            f" following files were provided (up to 10 shown):"
            f" {[a.file_name for a in analyses[:10]]}"
        )
    analysis = analyses[0]
    if analysis.analysis_spec["AnalysisType"] != AnalysisTypes.RULE:
        return (
            f"Only rules are supported for performance testing, but {analysis.analysis_spec['AnalysisType']}"
            f" was provided in {analysis.file_name}"
        )
    return analysis


def validate_log_type(
    log_type: Optional[str], rule: ClassifiedAnalysis
) -> Tuple[Optional[str], Optional[str]]:
    rule_log_types = rule.analysis_spec.get("LogTypes", [])
    if log_type is None:
        if len(rule_log_types) > 1:
            return (
                log_type,
                f"Multiple log types specified for {rule.file_name}: {rule_log_types}, please use the"
                f" --log-type arg to specify one.",
            )
        log_type = rule_log_types[0]
    else:
        if not str(log_type).casefold() in map(str.casefold, rule_log_types):
            return (
                log_type,
                f"Provided log type {log_type} was not found in log types for {rule.file_name}:"
                f" {rule_log_types}",
            )
    return log_type, None


def validate_hour(
    hour: Optional[datetime.datetime], log_type: str, backend: BackendClient
) -> Union[datetime.datetime, str]:
    now_truncated = datetime.datetime.now().replace(
        minute=0, second=0, microsecond=0
    ).astimezone() + datetime.timedelta(microseconds=-1)
    window_begin = now_truncated + datetime.timedelta(weeks=-2, microseconds=1)
    if hour is None:
        end_time = now_truncated
        start_time = window_begin
        err_msg = (
            f"No data found on Panther for log_type {log_type} in past two weeks. This can occur if"
            f" ingestion began within the last 24 hours."
        )
    else:
        if hour < window_begin:
            return f"Provided hour {hour.isoformat()} is too old. Please provide a time no older than {window_begin}"
        start_time = hour.replace(minute=0, second=0, microsecond=0)
        end_time = start_time + datetime.timedelta(hours=1, microseconds=-1)
        hour = start_time
        err_msg = (
            f"No data found on Panther for log_type {log_type} at specified hour: {hour.isoformat()}. Please"
            f" try another hour or leave the argument blank and one will be selected for you."
        )
    metrics_response = backend.get_metrics(
        MetricsParams(
            from_date=start_time,
            to_date=end_time,
            interval_in_minutes=60,
        )
    )
    data_for_log_type = next(
        (
            x
            for x in metrics_response.data.bytes_processed_per_source
            if x.label.casefold() == log_type.casefold()
        ),
        None,
    )
    if data_for_log_type is None:
        return err_msg

    if hour is not None and len(data_for_log_type.breakdown) > 1:
        return (
            "Internal error: time window too large. Please report this error to someone at Panther."
        )

    max_data_hour = max(
        data_for_log_type.breakdown, key=data_for_log_type.breakdown.get, default=None  # type: ignore
    )
    if max_data_hour is None or data_for_log_type.breakdown[max_data_hour] == 0:
        return err_msg
    hour = dateutil.parser.parse(max_data_hour)

    return hour


def generate_command_log_text(hour: datetime.datetime) -> List[str]:
    command = sys.argv[1:].copy()
    if "--hour" not in command:
        command.append("--hour")
        command.append(hour.isoformat())

    return [
        "To reproduce this benchmark on the same environment, please run:",
        "panther_analysis_tool " + " ".join(command),
    ]


def write_output(out: str, to_write: List[str], now: datetime.datetime) -> None:
    with open(out + f"/benchmark-{int(now.timestamp())}", "a", encoding="utf-8") as filename:
        to_write.insert(0, f"Writing to file: {filename.name}")
        log_and_write_to_file(to_write, filename)


def nanos_to_seconds(nanos: float) -> float:
    return datetime.timedelta(microseconds=nanos / 1000).total_seconds()


def log_output(
    out: str,
    hour: datetime.datetime,
    iterations: List[PerformanceTestIteration],
    rule: ClassifiedAnalysis,
    now: datetime.datetime,
) -> None:
    to_write = generate_command_log_text(hour)

    median_read_time_nanos = median([i.read_time_nanos for i in iterations])
    median_processing_time_nanos = median([i.processing_time_nanos for i in iterations])

    median_in_minutes = nanos_to_seconds(median_read_time_nanos + median_processing_time_nanos) / 60
    descriptor_string = "Less performant"
    if median_in_minutes < 1:
        descriptor_string = "Highly performant"
    elif median_in_minutes >= 10:
        descriptor_string = "At risk of timing out"

    to_write.extend(
        [
            "",
            f"Performance tested over {len(iterations)} iterations",
            f"Mean read time (seconds): {nanos_to_seconds(mean([i.read_time_nanos for i in iterations]))}",
            f"Median read time (seconds): {nanos_to_seconds(median_read_time_nanos)}",
            f"Max read time (seconds): {nanos_to_seconds(max([i.read_time_nanos for i in iterations]))}",  # pylint: disable=R1728
            f"Min read time (seconds): {nanos_to_seconds(min([i.read_time_nanos for i in iterations]))}",  # pylint: disable=R1728
            f"Mean processing time (seconds): {nanos_to_seconds(mean([i.processing_time_nanos for i in iterations]))}",
            f"Median processing time (seconds): {nanos_to_seconds(median_processing_time_nanos)}",
            f"Max processing time (seconds): {nanos_to_seconds(max([i.processing_time_nanos for i in iterations]))}",  # pylint: disable=R1728
            f"Min processing time (seconds): {nanos_to_seconds(min([i.processing_time_nanos for i in iterations]))}",  # pylint: disable=R1728
            "",
            "Detection performance ranges:",
            "< 1 minute: Highly performant",
            "1 minute to 10 minutes: Less performant, but unlikely to cause issues unless running alongside other less"
            " performant rules",
            "10+ minutes: At risk of timing out. Please improve performance",
            "",
            f"*** Rule {rule.file_name} is: {descriptor_string} ***",
            "",
            "Full record:",
        ]
    )
    to_write.extend([str(iteration) + "\n-----" for iteration in iterations])

    write_output(out, to_write, now)


def log_extreme_timeout(
    out: str,
    hour: datetime.datetime,
    now: datetime.datetime,
) -> None:
    to_write = generate_command_log_text(hour)
    to_write.append(
        "Your detection has timed out before completing the benchmark one hour of data! Please improve performance."
    )
    write_output(out, to_write, now)


def log_error(out: str, hour: datetime.datetime, now: datetime.datetime) -> None:
    to_write = generate_command_log_text(hour)
    to_write.append("benchmark failed with an error. Please ensure the correctness of your rule.")
    write_output(out, to_write, now)
